# -*- coding: utf-8 -*-

from collections import Counter
import itertools

from pyramid.httpexceptions import HTTPBadRequest
from pyramid.settings import asbool

from slashdb.exceptions import HTTPAmbiguousNames, ParserException
from slashdb.models import IModel
# from slashdb.utils import profile


# dictionary that keeps renderers accessible by extension or mime type
RENDERERS = {}
# dictionary that keeps error renderers accessible by extension or mime type
ERROR_RENDERERS = {}
# dictionary that keeps parser accessible by extension of mime type
PARSERS = {}

# dictionary that keeps renderers in list
# categorized by data structures that renderer is capable to produce
RENDERABLE = {
    'value_single': [],
    'value_many': [],
    'dictlike_single': [],
    'dictlike_many': []
}


class IResponseIterator(object):
    """Interface for SlashDB Renderer Emitter.

    When Renderer is streaming data using pyramid.response.Response(app_iter=output_iterator)
    the output_iterator should implement method output_iterator.close() that is called after last chunk.
    This is a right place for response cleanup like closing DataProxy session etc.

    """

    def close(self):
        raise NotImplementedError


class IRenderer(object):
    """Interface for SlashDB Renderers.

    Attributes extensions and mime_types define which renderer to be used when request comes.

    Attributes:
        extensions (tuple): tuple of strings of file extensions without dots like ('htm', 'html')
        mime_types (tuple): tuple of strings of mime types like ('text/html',)
        renderable (tuple): tuple of string of data structures renderer is able to produce
            (keys of slashdb.formats.RENDERABLE dict). For the moment it is used by html_sdb
            to figure out which url link to show.

    """

    extensions = tuple()
    mime_types = tuple()
    renderable = tuple()
    content_type = None

    @classmethod
    def error_response(cls, exc, request):
        raise NotImplementedError

    def __init__(self, context, data_proxy, request):
        self.context = context
        self.data_proxy = data_proxy
        self.request = request

    def is_streaming_response(self):
        """Should return true if response will send data in chunks.

        This value should be outcome of checking request.registry.settings (ini_settings).
        ini_settings.get('stream_by_default') - SlashDB default settings
        ini_settings.get('[plugin_name].stream_by_default') - SlashDB renderer plugin settings
        request.params.get('stream') - GET parameter provided in URL

        """
        raise NotImplementedError

    def render(self):
        """Method that creates output data for response.

        :return: output data as basestring or iterator
        """

        raise NotImplementedError

    def response_postprocessing(self, response):
        """Method for modifying freshly created response.

        Allows renderer to make some changes in response like add a header.

        :param response: created response object with defined body or app_iter, content type and charset
        :return: None
        """

        pass

    @staticmethod
    def _encode_explanation(explanation, output_encoding):
        u"""A workaround for Unicode errors for the Warning header explanation text."""
        try:
            exp = explanation.encode(output_encoding)
        except (UnicodeEncodeError, UnicodeDecodeError):
            exp = explanation

        exp = exp.strip()
        exp = exp.replace('\n', ' ')
        return exp


class DefaultIsStreamingMixin(object):
    u"""This mixin implements the default 'is_streaming_response' method for the IRenderer."""

    stream_by_default_str = ''

    def asbool(self, s):
        u"""A wrapper around asbool - usefull for implementing own logic for the Renderer."""
        return asbool(s)

    def is_streaming_response(self):
        """Checks url param, renderer settings and slashdb general settings."""
        url_param = self.request.params.get('stream')
        renderer_default = self.ini_settings.get(self.stream_by_default_str)
        slashdb_default = self.ini_settings.get('stream_by_default')

        if url_param == '':
            raise HTTPBadRequest(explanation="URL parameter stream is missing value (true or false).")
        elif url_param:
            return self.asbool(url_param)
        elif renderer_default:
            return self.asbool(renderer_default)
        elif slashdb_default:
            return self.asbool(slashdb_default)
        else:
            return False


class IParser(object):
    """Interface for SlashDB parsers.

    Attributes extensions and mime_types define which parser to use when request comes.

    Attributes:
        extensions (tuple): tuple of strings of file extensions without dots like ('htm', 'html')
        mime_types (tuple): tuple of strings of mime types like ('text/html',)

    """

    extensions = tuple()
    mime_types = tuple()

    def __init__(self, context, data_proxy, request):
        self.context = context
        self.data_proxy = data_proxy
        self.request = request

    def parse(self, data_str):
        """Returns list of dictionaries with data.

        Keys in dictionary must be IModel attribute names.
        Values in dictionary must be python types.

        Used by data_discovery_views.
        put_view uses only first element of the list for update
        post_view creates all IModel objects and saves them using slashdb.models.IDataProxy.

        """

        raise NotImplementedError


def bit2int(value, as_str=True):
    u"""Converts bits to int."""
    if '\\x' in repr(value) and isinstance(value, basestring):
        try:
            tmp = int(value.encode('hex'), 16)
        except Exception:
            pass
        else:
            return str(tmp) if as_str else tmp
    return value


# @profile
def adjust_bits(inp, bit2int=bit2int, sequence_types=(list, tuple)):
    u"""Clean-up all the bit mess in current data."""
    if isinstance(inp, IModel):
        return inp
    elif isinstance(inp, sequence_types) and len(inp):
        return tuple(adjust_bits(el) for el in inp)
    elif isinstance(inp, basestring):
        return bit2int(inp, False)
    elif isinstance(inp, dict):
        return {k: bit2int(v, False) for k, v in inp.iteritems()}
    return inp


def get_head_and_body(body):
    u"""Returns a initialized DataContainer."""
    if hasattr(body, '__iter__') and hasattr(body, 'next'):
        try:
            head = body.next()
        except StopIteration:
            head = None
        else:
            body = itertools.chain((head,), body)

        return head, body

    try:
        return body[0], body
    except (TypeError, KeyError):
        return body, body
    except IndexError:
        return None, body


def find_ambiguous(keys):
    """Finds repeated elements in iterable.

    :type keys: Iterable
    :param keys: an iterable of keys i.e. columns in an table

    :returns: list of items repeating in the provided keys iterable.
    :rtype: list

    """
    key_counter = Counter(keys)
    ambiguous = [key for key, count in key_counter.viewitems() if count > 1]

    return ambiguous


def check_for_ambiguous_keys(keys):
    u""""A helper function raising given exception when ambiguous have been found."""
    if keys:
        ambiguous = find_ambiguous(keys)
        if ambiguous:
            raise HTTPAmbiguousNames(ambiguous)


class DecoderToolbox(object):
    u"""A collection of common decoder tools."""

    supported_scalar_lengths = (0, 1)
    _format_parse_exception = None

    def __init__(self, format_parse_exception_fn):
        self._format_parse_exception = format_parse_exception_fn

    def check_scalar(self, scalars, key, *args):
        u"""Helper method - checks if there's a scalar provided and if it matches what we get form the incoming data."""
        if scalars and key != scalars[0]:
            raise self._format_parse_exception([key], *args)

    def check_columns(self, column_names, keys, *args):
        u"""Helper method - checks if we have invalid column names in the incoming keys."""
        invalid_headers = set(keys) - column_names
        if invalid_headers:
            raise self._format_parse_exception([h for h in keys if h in invalid_headers], *args)

    @staticmethod
    def get_parse_value_fn(get_python_value_fn):
        u"""Helper factory method - returns a parsing function."""
        def parse_value(k, v):
            try:
                pv = get_python_value_fn(k, v)
            except ValueError as exc:
                raise ParserException(exc.message)
            return pv
        return parse_value
