# -*- coding: utf-8 -*-

"""Package with CSV renderer and parser for SlashDB."""
import csv
import logging
import StringIO

from pyramid.httpexceptions import HTTPBadRequest, HTTPNotImplemented
from pyramid.response import Response
from pyramid.settings import asbool

from slashdb.formats import DefaultIsStreamingMixin, IParser, IRenderer, check_for_ambiguous_keys
from slashdb.formats.csv_sdb.decoder import CSVDecoder
from slashdb.formats.csv_sdb.emitter import CSVEmitter
from slashdb.formats.csv_sdb.encoder import CSVEncoder
# from slashdb.prof import profile
from slashdb.querydef_utils import make_query_url_template

__version__ = '0.0.1'

log = logging.getLogger('slashdb')

EXTENSIONS = ('csv', )
MIME_TYPES = ('text/csv',)


truthy = frozenset(('t', 'true', 'y', 'yes', 'on', '1'))
falsy = frozenset(('f', 'false', 'n', 'no', 'off', '0'))


class CSVRenderer(DefaultIsStreamingMixin, IRenderer):

    extensions = EXTENSIONS
    mime_types = MIME_TYPES
    renderable = ('value_single', 'value_many', 'dictlike_single', 'dictlike_many')
    content_type = mime_types[0]
    stream_by_default_str = 'csv_sdb.stream_by_default'

    @classmethod
    def error_response(cls, exc, request):
        ini_settings = request.registry.settings
        sdb_settings = ini_settings['sdb_settings']
        output_encoding = sdb_settings['output_encoding']
        data = {
            'http_code': exc.status_code,
            'description': exc.explanation
        }

        try:
            query_id = request.matchdict['query_id']
            query_config = sdb_settings['config_storage'].querydefs[query_id]
            data['url_template'] = make_query_url_template(query_config, 'csv')
        except (AttributeError, TypeError, KeyError):
            pass  # no change to data dict

        sio = StringIO.StringIO()
        writer = csv.DictWriter(sio, fieldnames=data.keys())
        writer.writeheader()
        writer.writerow(data)
        body = sio.getvalue()

        explanation = cls._encode_explanation(exc.explanation, output_encoding)
        response = Response(
            body=body,
            headers=[('Warning', explanation)],
            status_int=exc.code
        )
        response.content_type = cls.mime_types[0]
        response.charset = output_encoding

        return response

    def __init__(self, context, data_proxy, request):
        IRenderer.__init__(self, context, data_proxy, request)

        self.db_encoding = context.db_config.db_encoding
        self.res_req_meta = data_proxy.res_req_meta
        self.depth = self.res_req_meta[-1].depth if self.res_req_meta else 0
        self.ini_settings = request.registry.settings
        self.output_encoding = self.ini_settings['output_encoding']

    @property
    def with_headers(self):
        """Checks if CSV output should contain headers in first line.

         First url param then renderer settings.

        """
        url_param = self.request.params.get('headers')
        renderer_default = self.ini_settings.get('csv_sdb.headers_by_default')

        def catch_bool_parse_exception(v):
            u"""Wrapper around self.asbool."""
            try:
                return self.asbool(v)
            except ValueError:
                raise HTTPBadRequest(explanation='Headers has to be True or False.')

        if url_param == '':
            raise HTTPBadRequest(explanation="URL parameter headers is missing value (true or false).")
        elif url_param:
            return catch_bool_parse_exception(url_param)
        elif renderer_default:
            return catch_bool_parse_exception(renderer_default)
        else:
            return True

    def asbool(self, s):
        """ Return the boolean value ``True`` if the case-lowered value of string
        input ``s`` is any of ``t``, ``true``, ``y``, ``on``, or ``1``, otherwise
        return the boolean value ``False``.  If ``s`` is the value ``None``,
        return ``False``.  If ``s`` is already one of the boolean values ``True``
        or ``False``, return it."""
        if s is None:
            return False
        if isinstance(s, bool):
            return s
        s = str(s).strip().lower()
        if s in truthy:
            return True
        elif s in falsy:
            return False
        raise ValueError()

    # @profile
    def render(self):
        if self.res_req_meta:
            check_for_ambiguous_keys(self.res_req_meta[-1].scalars)

        if self.is_streaming_response():
            if self.depth > 0:
                raise HTTPNotImplemented(
                    explanation="Streamed rendering to CSV of nested resources is not implemented."
                )

            return CSVEmitter(self.data_proxy, self.with_headers, self.db_encoding, self.output_encoding)
        else:
            csv_encoder = CSVEncoder(
                self.data_proxy, self.with_headers, self.db_encoding, self.output_encoding, flatten=self.depth > 0
            )
            return csv_encoder.encode()


class CSVParser(IParser):

    extensions = EXTENSIONS
    mime_types = MIME_TYPES

    def __init__(self, context, data_proxy, request):
        IParser.__init__(self, context, data_proxy, request)

        self.db_name = context.db_id
        self.last_res_req_meta = data_proxy.res_req_meta[-1]
        self.csv_null_str = request.params.get(
            'csvNullStr', request.registry.settings.get('csv_sdb.csv_null_str', None)
        )
        self.input_encoding = request.charset

    def parse(self, data_str):
        csv_decoder = CSVDecoder(self.db_name, self.last_res_req_meta, self.csv_null_str, self.input_encoding)
        update_dict = csv_decoder.decode(data_str)
        return update_dict


Renderer = CSVRenderer
Parser = CSVParser
