# -*- coding: utf-8 -*-

"""Package with JSON renderer and parser for SlashDB."""
import json

from pyramid.response import Response
from pyramid.settings import asbool

from slashdb.formats import DefaultIsStreamingMixin, IParser, IRenderer, check_for_ambiguous_keys
from slashdb.formats.json_sdb.decoder import JSONDecoder
from slashdb.formats.json_sdb.emitter import JSONEmitter
# from slashdb.prof import profile
from slashdb.querydef_utils import make_query_url_template

__version__ = '0.0.1'

EXTENSIONS = ('json', )
MIME_TYPES = ('application/json', )


class JSONConfig(object):
    u"""JSON config info container."""

    __slots__ = ('indent', 'minimize', 'output_encoding')

    def __init__(self, ini_settings=None, indent=None, minimize=False, output_encoding='utf-8'):
        ini_settings = ini_settings or {}
        self.indent = indent or ini_settings.get('json_sdb.indent', None)
        self.minimize = minimize or asbool(ini_settings.get('json_sdb.minimize', False))
        self.output_encoding = output_encoding or ini_settings.get('output_encoding', 'utf-8')

        if self.minimize:
            self.indent = None

        if self.indent is not None:
            self.indent = ' ' * int(self.indent)


class JSONRenderer(DefaultIsStreamingMixin, IRenderer):
    """JSON renderer for SlashDB."""

    extensions = EXTENSIONS
    mime_types = MIME_TYPES
    renderable = ('value_single', 'value_many', 'dictlike_single', 'dictlike_many')
    content_type = mime_types[0]
    stream_by_default_str = 'json_sdb.stream_by_default'

    @classmethod
    def error_response(cls, exc, request):
        sdb_settings = request.registry.settings['sdb_settings']
        output_encoding = sdb_settings['output_encoding']
        data = {
            'http_code': exc.status_code,
            'description': exc.explanation
        }

        try:
            query_id = request.matchdict['query_id']
            query_config = sdb_settings['config_storage'].querydefs[query_id]
            data['url_template'] = make_query_url_template(query_config, 'json')
        except (AttributeError, TypeError, KeyError):
            pass  # no change to data dict

        body = json.dumps(data)
        explanation = cls._encode_explanation(exc.explanation, output_encoding)
        response = Response(
            body=body,
            headers=[('Warning', explanation)],
            status_int=exc.code
        )
        response.content_type = cls.content_type
        response.charset = output_encoding

        return response

    def __init__(self, context, data_proxy, request):
        IRenderer.__init__(self, context, data_proxy, request)

        self.db_name = context.db_id
        self.db_config = context.db_config
        self.db_encoding = context.db_config.db_encoding
        self.query_id = context.query_id
        self.slash_escaping_fun = context.get_slash_escaping_fun()

        self.res_req_meta = data_proxy.res_req_meta

        self.host_url = request.host_url
        self.depth = self.res_req_meta[-1].depth if self.res_req_meta else 0
        self.ini_settings = request.registry.settings
        self.json_config = JSONConfig(self.ini_settings)

    # @profile
    def render(self):
        if self.res_req_meta:
            check_for_ambiguous_keys(self.res_req_meta[-1].scalars)

        is_streaming = self.is_streaming_response()
        emtr = JSONEmitter(
            self.data_proxy,
            self.db_encoding,
            self.json_config,
            self.depth,
            self.slash_escaping_fun,
            is_streaming
        )

        if is_streaming:
            return emtr

        return emtr._to_string()

    def response_postprocessing(self, response):
        pass


class JSONParser(IParser):

    extensions = EXTENSIONS
    mime_types = MIME_TYPES

    def __init__(self, context, data_proxy, request):
        IParser.__init__(self, context, data_proxy, request)

        self.res_req_meta = data_proxy.res_req_meta
        self.input_encoding = request.charset

    def parse(self, data_str):
        json_decoder = JSONDecoder(self.res_req_meta, self.input_encoding)
        update_dict = json_decoder.decode(data_str)
        return update_dict


Renderer = JSONRenderer
Parser = JSONParser
