# -*- coding: utf-8 -*-

"""XML renderer and parser for SlashDB.

Module provides class of SlashDB IRenderer capable of returning pyramid response with SlashDB resources encoded in XML.


Renderer is configurable in settings with parameters:

xml_sdb.root_tag - sets root tag in xml, by default: SlashDB
xml_sdb.xmlns - namespace for XML instance, by default: http://www.vtenterprise.com/slashdb

"""

import logging

from pyramid.httpexceptions import HTTPBadRequest
from pyramid.response import Response
from pyramid.settings import asbool

from slashdb.formats import DefaultIsStreamingMixin, IParser, IRenderer, check_for_ambiguous_keys
from slashdb.formats.xml_sdb.decoder import XMLDecoder
from slashdb.formats.xml_sdb.emitter import XMLEmitter, XMLRecordsetEmitter
from slashdb.formats.xml_sdb.utils import querydef_xml_error_template, xml_error_template
from slashdb.formats.xsd_sdb import XSDEncoder
# from slashdb.prof import profile
from slashdb.querydef_utils import make_query_url_template

__version__ = '0.0.1'

log = logging.getLogger('slashdb')

EXTENSIONS = ('xml', )
MIME_TYPES = ('text/xml', 'application/xml')


class XMLRenderer(DefaultIsStreamingMixin, IRenderer):

    extensions = EXTENSIONS
    mime_types = MIME_TYPES
    renderable = ('value_single', 'value_many', 'dictlike_single', 'dictlike_many')
    content_type = mime_types[0]
    stream_by_default_str = 'xml_sdb.stream_by_default'

    @classmethod
    def error_response(cls, exc, request):
        ini_settings = request.registry.settings
        sdb_settings = ini_settings['sdb_settings']
        output_encoding = sdb_settings['output_encoding']
        data = {
            'root_tag': ini_settings.get('xml_sdb.root_tag', 'SlashDB'),
            'http_code': exc.status_code,
            'description': exc.explanation
        }

        try:
            query_id = request.matchdict['query_id']
            query_config = sdb_settings['config_storage'].querydefs[query_id]
        except (AttributeError, TypeError, KeyError):
            body = xml_error_template % data
        else:
            data['url_template'] = make_query_url_template(query_config, 'xml')
            body = querydef_xml_error_template % data

        explanation = cls._encode_explanation(exc.explanation, output_encoding)
        response = Response(
            body=body,
            headers=[('Warning', explanation)],
            status_int=exc.code
        )
        response.content_type = cls.content_type
        response.charset = output_encoding

        return response

    def __init__(self, context, data_proxy, request):
        IRenderer.__init__(self, context, data_proxy, request)

        self.ini_settings = request.registry.settings
        self.res_req_meta = data_proxy.res_req_meta
        self.depth = self.res_req_meta[-1].depth if self.res_req_meta else 0
        self.output_encoding = self.ini_settings.get('output_encoding', 'utf-8')
        self.slash_escaping_fun = context.get_slash_escaping_fun()
        self.xml_type = self.request.params.get('xmlType')

    @property
    def nil_visible(self):
        url_param = self.request.params.get('nil_visible')
        renderer_default = self.ini_settings.get('xml_sdb.nil_visible')

        if url_param == '':
            raise HTTPBadRequest(explanation="URL parameter nil_visible is missing value (true or false).")
        elif url_param:
            return asbool(url_param)
        elif renderer_default:
            return asbool(renderer_default)
        else:
            return False

    # @profile
    def render(self):
        if self.res_req_meta:
            check_for_ambiguous_keys(self.res_req_meta[-1].scalars)

        emtr = XMLEmitter
        is_streaming = self.is_streaming_response()

        if self.context.query_id and self.xml_type == 'adPersistXML':
            emtr = XMLRecordsetEmitter

        xml_gen = emtr(
            self.context,
            self.data_proxy,
            self.request.host_url,
            self.depth,
            self.ini_settings,
            self.nil_visible,
            self.slash_escaping_fun,
            is_streaming
        )

        if is_streaming:
            return xml_gen

        # profile.dump_stats("xml_emitter_opti1.fprof")
        return xml_gen._to_string()


class XMLParser(IParser):

    extensions = EXTENSIONS
    mime_types = MIME_TYPES

    def __init__(self, context, data_proxy, request):
        IParser.__init__(self, context, data_proxy, request)

        self.url_segments = context.url_segments
        self.db_name = context.db_id
        self.settings = context.settings
        self.query_id = context.query_id

        self.res_req_meta = data_proxy.res_req_meta
        self.host_url = request.host_url
        self.req_params = request.params
        self.ini_settings = request.registry.settings

    def parse(self, data_str):
        xsd_schema = XSDEncoder(
            self.query_id,
            self.data_proxy,
            int(self.req_params.get('depth', 0)),
            self.req_params.get('cardinality', 'unbounded'),
            self.ini_settings
        ).encode()

        xml_decoder = XMLDecoder(
            self.url_segments,
            self.res_req_meta,
            self.db_name,
            self.settings,
            self.host_url,
            self.req_params,
            xsd_schema
        )

        update_dict = xml_decoder.decode(data_str)
        return update_dict


Renderer = XMLRenderer
Parser = XMLParser
