# -*- coding: utf-8 -*-

"""Package with XSD renderer for SlashDB."""

import os
import urllib
from xml.sax import saxutils

from pyramid.httpexceptions import HTTPBadRequest, HTTPNotImplemented, HTTPSeeOther
from pyramid.response import Response

from slashdb.formats import DefaultIsStreamingMixin, IRenderer, check_for_ambiguous_keys
from slashdb.formats.xsd_sdb.encoder import XSDEncoder
from slashdb.formats.xsd_sdb.utils import get_schema_url, querydef_xsd_error_template, xsd_error_template
# from slashdb.prof import profile
from slashdb.querydef_utils import make_query_url_template

__version__ = '0.0.1'

EXTENSIONS = ('xsd', )
MIME_TYPES = ('text/xsd', 'application/xsd')


class XSDRenderer(DefaultIsStreamingMixin, IRenderer):

    extensions = EXTENSIONS
    mime_types = MIME_TYPES
    renderable = ('value_single', 'value_many', 'dictlike_single', 'dictlike_many')
    content_type = 'text/xml'
    stream_by_default_str = 'xsd_sdb.stream_by_default'

    @classmethod
    def error_response(cls, exc, request):
        ini_settings = request.registry.settings
        sdb_settings = ini_settings['sdb_settings']
        output_encoding = sdb_settings['output_encoding']
        data = {
            'root_tag': ini_settings.get('xml_sdb.root_tag', 'SlashDB'),
            'http_code': exc.status_code,
            'description': exc.explanation
        }

        try:
            query_id = request.matchdict['query_id']
            query_config = sdb_settings['config_storage'].querydefs[query_id]
        except (AttributeError, TypeError, KeyError):
            body = xsd_error_template % data
        else:
            data['url_template'] = make_query_url_template(query_config, 'xsd')
            body = querydef_xsd_error_template % data

        explanation = cls._encode_explanation(exc.explanation, output_encoding)
        response = Response(
            body=body,
            headers=[('Warning', explanation)],
            status_int=exc.code
        )
        response.content_type = cls.content_type
        response.charset = output_encoding

        return response

    def __init__(self, context, data_proxy, request):
        IRenderer.__init__(self, context, data_proxy, request)

        self.url_segments = context.url_segments
        self.accept_header = context.accept_header
        self.db_id = context.db_id
        self.query_id = context.query_id
        self.settings = context.settings

        self.res_req_meta = data_proxy.res_req_meta
        self.is_single_resource = data_proxy.is_single_resource

        self.host = request.host
        self.host_url = request.host_url
        self.ini_settings = request.registry.settings

        self.req_params = request.params
        self.depth = self.res_req_meta[-1].depth if self.res_req_meta else 0
        self.cardinality = self.request.params.get('cardinality', 'unbounded')
        self.output_encoding = self.ini_settings.get('output_encoding', 'utf-8')

    def _get_schema_url(self):
        schema_url = ''
        schema_params = ''

        if self.query_id:
            ext = u'' if self.accept_header else u'.xsd'
            schema_url = u'{}/query/{}/{}{}'.format(self.host_url, self.query_id, os.path.join(*self.url_segments), ext)
            schema_url = saxutils.escape(schema_url.encode(self.output_encoding))
            schema_params = ''
        elif self.res_req_meta:
            res_req_meta = self.res_req_meta[-1]
            schema_url, schema_params = get_schema_url(
                res_req_meta.resource_class,
                res_req_meta.scalars,
                self.is_single_resource,
                self.host_url,
                self.db_id,
                self.depth,
                accept_header=self.accept_header
            )
            schema_url = schema_url.encode(self.output_encoding)
            schema_params = schema_params.encode(self.output_encoding)

        return schema_url, schema_params

    # @profile
    def render(self):
        if self.res_req_meta:
            check_for_ambiguous_keys(self.res_req_meta[-1].scalars)

        if self.is_streaming_response():
            raise HTTPNotImplemented(explanation="Streaming XSD is not implemented.")

        xsd_maker = XSDEncoder(
            self.query_id,
            self.data_proxy,
            self.depth,
            self.cardinality,
            self.ini_settings
        )

        if self.query_id:
            return xsd_maker.encode_sqlpassthru()
        elif self.res_req_meta:
            schema_url, schema_params = self._get_schema_url()
            if urllib.unquote(schema_url) != urllib.unquote(self.request.path_url):
                raise HTTPSeeOther(location=schema_url + schema_params)

            return xsd_maker.encode()

        raise HTTPBadRequest(explanation="XSD is available only for Data Discovery requests.")

    def response_postprocessing(self, response):
        pass


Renderer = XSDRenderer
Parser = None
